/*
 * rwblock.c
 * Blocking read/write utility.
 *
 * Author: Phil Burk, http://www.softsynth.com
 *
 * This program uses the PortAudio Portable Audio Library.
 * For more information see: http://www.audiomulch.com/portaudio/
 * Copyright (c) 1999-2000 Ross Bencina and Phil Burk
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files
 * (the "Software"), to deal in the Software without restriction,
 * including without limitation the rights to use, copy, modify, merge,
 * publish, distribute, sublicense, and/or sell copies of the Software,
 * and to permit persons to whom the Software is furnished to do so,
 * subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * Any person wishing to distribute modifications to the Software is
 * requested to send the modifications to the original developer so that
 * they can be incorporated into the canonical version.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR
 * ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "portaudio.h"
#include "ringbuffer.h"
#include "rwblock.h"
#include <string.h>
static int blockingIOCallback(	void *inputBuffer, void *outputBuffer,
						unsigned long framesPerBuffer,
				        PaTimestamp outTime, void *userData );
static PaError RWBL_InitFIFO( RingBuffer *rbuf, int numFrames, int bytesPerFrame );
static PaError RWBL_TermFIFO( RingBuffer *rbuf );
/* Called from PortAudio.
 * Read and write data only there is room in FIFOs.
 */
static int blockingIOCallback(	void *inputBuffer, void *outputBuffer,
						unsigned long framesPerBuffer,
				        PaTimestamp outTime, void *userData )
{
	RWBL *data = (RWBL*)userData;
	int numBytes = data->bytesPerFrame * framesPerBuffer;
	(void) outTime;
	
/* This may get called with NULL inputBuffer during initial setup. */
	if( inputBuffer != NULL )
	{
		RingBuffer_Write( &data->inFIFO, inputBuffer, numBytes );
	}
	if( outputBuffer != NULL )
	{
		RingBuffer_Read( &data->outFIFO, outputBuffer, numBytes );
	}
	
	return 0;
}
/* Allocate buffer. */
static PaError RWBL_InitFIFO( RingBuffer *rbuf, int numFrames, int bytesPerFrame )
{
	int numBytes = numFrames * bytesPerFrame;
	char *buffer = (char *) malloc( numBytes );
	if( buffer == NULL ) return paInsufficientMemory;
	return (PaError) RingBuffer_Init( rbuf, numBytes, buffer );
}
/* Free buffer. */
static PaError RWBL_TermFIFO( RingBuffer *rbuf )
{
	if( rbuf->buffer ) free( rbuf->buffer );
	rbuf->buffer = NULL;
	return paNoError;
}
/************************************************************
 * Write data to ring buffer.
 * Will not return until all the data has been written.
 */
int WriteAudioStream( RWBL *rwbl, void *data, int numFrames )
{
	int bytesWritten;
	int numBytes = rwbl->bytesPerFrame * numFrames;
	while( numBytes > 0)
	{
		bytesWritten = RingBuffer_Write( &rwbl->outFIFO, data, numBytes );
		numBytes -= bytesWritten;
		if( numBytes > 0) Pa_Sleep(10);
	}
	return numFrames;
}
/************************************************************
 * Read data from ring buffer.
 * Will not return until all the data has been read.
 */
int ReadAudioStream( RWBL *rwbl, void *data, int numFrames )
{
	int bytesRead;
	int numBytes = rwbl->bytesPerFrame * numFrames;
	while( numBytes > 0)
	{
		bytesRead = RingBuffer_Read( &rwbl->inFIFO, data, numBytes );
		numBytes -= bytesRead;
		if( numBytes > 0) Pa_Sleep(10);
	}
	return numFrames;
}
static unsigned int RoundUpToNextPowerOf2( unsigned int n )
{
	int numBits = 0;
	if( ((n-1) & n) == 0) return n; /* Already Power of two. */
	while( n > 0 )
	{
		n= n>>1;
		numBits++;
	}
	return (1<<numBits);
}
/************************************************************
 * Opens a PortAudio stream with default characteristics.
 * Allocates RWBL structure.
 *
 * flags parameter can be an ORed combination of:
 *    RWBL_READ, RWBL_WRITE, and either RWBL_MONO or RWBL_STEREO
 */
#define FRAMES_PER_BUFFER    (256)
PaError OpenAudioStream( RWBL **rwblPtr, double sampleRate,
				  PaSampleFormat format, int flags )
{
	int   bytesPerSample;
	int   doRead = 0;
	int   doWrite = 0;
	PaError err;
	RWBL *rwbl;
	int   minNumBuffers;
	int   numFrames;
/* Allocate RWBL structure for caller. */
	rwbl = (RWBL *) malloc( sizeof(RWBL) );
	if( rwbl == NULL ) return paInsufficientMemory;
	memset( rwbl, 0, sizeof(RWBL) );
/* Determine size of a sample. */
	switch(format )
	{
	case paInt16:
		bytesPerSample = 2;
		break;
	case paFloat32:
	case paInt32:
	default:
		bytesPerSample = 4;
		break;
	}
	rwbl->samplesPerFrame = ((flags&RWBL_MONO) != 0) ? 1 : 2;
	rwbl->bytesPerFrame = bytesPerSample * rwbl->samplesPerFrame;
/* Initialize PortAudio  */
	err = Pa_Initialize();
	if( err != paNoError ) goto error;
 /* Warning: numFrames must be larger than amount of data processed per interrupt
  *    inside PA to prevent glitches. Just to be safe, adjust size upwards.
  */
	minNumBuffers = Pa_GetMinNumBuffers( FRAMES_PER_BUFFER, sampleRate );
	numFrames = minNumBuffers * FRAMES_PER_BUFFER;
	numFrames = RoundUpToNextPowerOf2( numFrames );
/* Initialize Ring Buffers */
	doRead = ((flags & RWBL_READ) != 0);
	doWrite = ((flags & RWBL_WRITE) != 0);
	if(doRead)
	{
		err = RWBL_InitFIFO( &rwbl->inFIFO, numFrames, rwbl->samplesPerFrame * bytesPerSample );
		if( err != paNoError ) goto error;
	}
	if(doWrite)
	{
		err = RWBL_InitFIFO( &rwbl->outFIFO, numFrames, rwbl->samplesPerFrame * bytesPerSample );
		if( err != paNoError ) goto error;
	}
	err = Pa_OpenStream(
		&rwbl->stream,
		(doRead ? Pa_GetDefaultInputDeviceID() : paNoDevice),
		(doRead ? rwbl->samplesPerFrame : 0 ),
		format,	
		NULL,
		(doWrite ? Pa_GetDefaultOutputDeviceID() : paNoDevice),
		(doWrite ? rwbl->samplesPerFrame : 0 ),
		format, 
		NULL,
		sampleRate,
		FRAMES_PER_BUFFER,
		minNumBuffers,
		paClipOff,       /* we won't output out of range samples so don't bother clipping them */
		blockingIOCallback,
		rwbl );
	if( err != paNoError ) goto error;
	
	err = Pa_StartStream( rwbl->stream );
	if( err != paNoError ) goto error;
	
	*rwblPtr = rwbl;
	return paNoError;
error:
	CloseAudioStream( rwbl );
	*rwblPtr = NULL;
	return err;
}
PaError CloseAudioStream( RWBL *rwbl )
{
	PaError err;
	err = Pa_CloseStream( rwbl->stream );
	if( err != paNoError ) goto error;
	Pa_Terminate();
error:
	RWBL_TermFIFO( &rwbl->inFIFO );
	RWBL_TermFIFO( &rwbl->outFIFO );
	return err;
}
